#include <Rcpp.h>
#include <iostream>

using namespace std;
using namespace Rcpp;

//---------------------------- helper -----------------------------------//
// ------------- basic -------------------//
// [[Rcpp::export]]
NumericVector rowSumsC(NumericMatrix x){
  int nrow = x.nrow(), ncol = x.ncol();
  NumericVector out(nrow);
  for (int i = 0; i < nrow; i++) {
    double total = 0;
    for (int j = 0; j < ncol; j++) {
      total += x(i, j);
    }
    out[i] = total;
  }
  return out;
}

// [[Rcpp::export]]
NumericVector colSumsC(NumericMatrix x){
  int nrow = x.nrow(), ncol = x.ncol();
  NumericVector out(ncol);
  for (int i = 0; i < ncol; i++) {
    double total = 0;
    for (int j = 0; j < nrow; j++) {
      total += x(j, i);
    }
    out[i] = total;
  }
  return out;
}

// [[Rcpp::export]]
NumericMatrix MatMul(NumericMatrix x, NumericMatrix y){
  int nrow = x.nrow(), ncol = y.ncol(), rank = x.ncol();
  NumericMatrix xy(nrow, ncol);
  for(int i = 0; i < nrow; i ++){
    for(int j = 0; j < ncol; j++){
      double temp = 0;
      for(int k = 0; k < rank; k ++){
        temp += x(i, k) * y(k, j);
      }
      xy(i, j) = temp;
    }
  }
  return xy;
}

//[[Rcpp::export]]
NumericMatrix NormByCol(NumericMatrix x){
  int nrow = x.nrow(), ncol = x.ncol();
  NumericMatrix out(nrow, ncol);
  NumericVector colsum = colSumsC(x);
  for (int i = 0; i < nrow; i++) {
    for (int j = 0; j < ncol; j++) {
       out(i, j) = x(i, j)/colsum[j];
    }
  }
  return out;
}

//[[Rcpp::export]]
NumericMatrix Transpose(NumericMatrix x){
  int nrow = x.nrow(), ncol = x.ncol();
  NumericMatrix out(ncol, nrow);
  for(int i = 0; i < nrow; i++){
    for(int j = 0; j < ncol; j++){
      out(j, i) = x(i, j);
    }
  }
  return out;
}



// ---------------- useful function ----------------- //
// [[Rcpp::export]]
void UpdateC(NumericMatrix& W, NumericMatrix& H, NumericMatrix V, NumericMatrix Cset, float alpha){
  //basic info
  int n_row = V.nrow(), n_col = V.ncol(), rank = W.ncol();
  // intermediate parameters
  double sum_H = 0, sum_W = 0, sum_product = 0, Cpar = 0, delta = 1e-20;//control the divider
  NumericVector newH(n_col), newW(n_row);
  
  //update
  for(int i = 0; i < rank; i++){
    //update H
    for(int j = 0; j < n_col; j++){
      sum_H = 0;
      sum_W = 0;
      
      for(int k = 0; k < n_row; k++){
        sum_W += W(k, i);
      }
      
      for(int k = 0; k < n_row; k++){
        sum_product = 0;
        for(int n = 0; n < rank; n++){
          sum_product += W(k, n) * H(n, j);
        }
        sum_H += V(k, j) * W(k, i)/std::max(sum_product, delta);
      }
      
      newH[j] = sum_H * (H(i, j)/std::max(sum_W, delta));
    }
    
    for(int j = 0;j < n_col; j++){
      H(i, j) = newH[j];
    }
    
    //update W
    for(int j = 0; j < n_row; j++){
      sum_H = 0;
      sum_W = 0;
      Cpar = 0;
      
      for(int k = 0; k < n_row; k++){
        if(Cset(j, k) == 1){
          Cpar += W(k, i);
        }
      }
      
      for(int k = 0; k < n_col; k++){
        sum_H += H(i, k);
      }
      
      for(int k = 0; k < n_col; k++){
        sum_product = 0;
        for(int n = 0; n < rank; n++){
          sum_product += W(j, n) * H(n, k);
        }
        sum_W += V(j, k) * H(i, k)/std::max(sum_product, delta);
      }
      
      newW[j] = sum_W * W(j, i)/std::max((sum_H + 2 * alpha * Cpar), delta);
    }
    
    // sum_W=0;
    // for(int j=0;j<n_row;j++){
    //     sum_W=sum_W+newW[j];
    // }
    for(int j = 0; j < n_row; j++){
      W(j, i) = newW[j];  // /sum_W;
    }
    // for(int j=0;j<n_col;j++){
    //    H[i][j]=H[i][j]*sum_W;
    // }
  }
  
  return;
  //List::create(_["W"] = W, _["H"] = H); //need to modify corresponding syntax in helper.R
}

// [[Rcpp::export]]
double CpenaltyC(NumericMatrix W, NumericMatrix H, NumericMatrix V, NumericMatrix Cset, double alpha){
  //basic info
  int n_row = V.nrow(), n_col = V.ncol(), rank = W.ncol();
  //double delta = 1e-20; //control the divider
  
  //U=W%*%t(W)
  NumericMatrix U(n_row, n_row);
  for(int i=0; i < n_row; i++){
    for(int j = 0; j < n_row; j++){
      U(i, j) = 0;
      for(int k = 0; k < rank; k++){
        U(i, j) += W(i, k) * W(j, k);
      }
    }
  }
  
  //WH=W%*%H
  NumericMatrix WH(n_row, n_col);
  for(int i = 0; i < n_row; i++){
    for(int j = 0; j < n_col; j++){
      WH(i, j) = 0;
      for(int k = 0; k < rank; k++){
        WH(i, j) += W(i, k) * H(k, j);
      }
    }
  }
  
  // double cpenalty=0, Psum=0;
  // for(int i = 0; i < n_row; i++){
  //   Psum=0;
  //   for(int j = 0; j < n_col; j++){
  //     if(V(i, j) > delta && WH(i, j) > delta){
  //       Psum += std::max(V(i, j) * (log(V(i, j)/WH(i, j)) - 1) + WH(i, j), delta);
  //     }else{
  //       Psum += abs(WH(i, j) - V(i, j));
  //     }
  //     if(Cset(i, j) == 1){
  //       Psum += alpha * U(i, j);
  //     }
  //   }
  //   cpenalty += Psum;
  // }
  
  double cpenalty=0, Psum=0;
  for(int i = 0; i < n_row; i++){
    Psum=0;
    for(int j = 0; j < n_col; j++){
      if(V(i, j) > 0 && WH(i, j) > 0){
        Psum += V(i, j) * log(V(i, j)/WH(i, j)) - V(i,j) + WH(i, j);
      }else{
        Psum += abs(WH(i, j) - V(i, j));
      }
      if(Cset(i, j) == 1){
        Psum += alpha * U(i, j);
      }
    }
    cpenalty += Psum;
  }
  
  
  return cpenalty;
}

// [[Rcpp::export]]
List ncNMFC(NumericMatrix V, int rank, NumericMatrix cset, double alpha0 = 10, int iter = 300, double eps = 0.0005){
  // Returns the NMF result
  //  Args:
  //   V: the matrix NMF to be applied on
  //   rank: the NMF rank
  //   alpha0: the coefficient of L1 penalty
  //   cset: a table containing all the pairs of conflicting bins
  //   iter: the maximal number of iterations
  //   eps: the convergence accuracy
  //   code_path: the path to NMF codes implemented in C++
  
  
  // initial the matrix
  // randomly set intial matrix
  int n_row = V.nrow(), n_col = V.ncol();
  double mean_V = mean(V);
  NumericMatrix W(n_row, rank, (abs(rnorm(n_row * rank) * sqrt(1/double(n_row)) + 1/double(n_row))).begin());
  W = NumericMatrix(n_row, rank, (NormByCol(W)).begin());
  NumericMatrix H(rank, n_col, (abs(rnorm(rank * n_col) * sqrt(mean_V * double(n_row)) + mean_V * double(n_row))).begin());
  
  
  // Code dependent alpha
  double alpha = mean(MatMul(W, H))/mean(MatMul(W, Transpose(W))) * alpha0;
  // initial error
  double err0 = CpenaltyC(W, H, V, cset, alpha), err = err0/2;
  int count = 0;
  
  while(abs(err - err0) >= eps * err0& count < iter){
    alpha = mean(MatMul(W, H))/mean(MatMul(W, Transpose(W))) * alpha0;
    count += 1;
    
    UpdateC(W, H, V, cset, alpha);
    //W = NumericMatrix(n_row, rank, (WH_result["W"]).begin());
    //H = NumericMatrix(rank, n_col, (WH_result["H"]).begin());
    if(count % 10 == 0){
      err0 = err;
      err = CpenaltyC(W, H, V, cset, alpha);
    }
  }
  
  return List::create(_["basis"] = W, _["coefficient"] = H, _["error"] = err);
}

